# scrapyd_manager/scrapyd_api.py
import requests
from django.utils import timezone
from typing import List
from logging import getLogger
from .cache import django_ttl_cache
from . import models
from django.db.models import Q
from typing import Protocol, Iterable
from datetime import datetime
from django.conf import settings


class SpiderGroupLike(Protocol):
    name: str
    code: str | None
    kwargs: dict
    settings: dict
    resolved_spiders: Iterable[models.Spider]


logger = getLogger(__name__)


class ScrapydResponseError(Exception):
    pass


class NodesSyncError(Exception):
    pass


def _auth_for_node(node: models.Node):
    """返回 node 的认证信息"""
    if getattr(node, "auth", False):
        return node.username, node.password
    return None


def start_spider(spider: models.Spider) -> str:
    """启动爬虫并返回 Job"""
    url = f"{spider.version.project.node.url}/schedule.json"
    kwargs = spider.kwargs.copy()
    for k, v in list(kwargs.items()):
        if k.startswith("__"):
            kwargs.pop(k)
    data = {
        "project": spider.version.project.name,
        "spider": spider.name,
        "setting": [f"{k}={v}" for k, v in spider.settings.items()],
        "jobid": spider.job_id,
        **kwargs,
    }
    resp = requests.post(url, data=data, auth=_auth_for_node(spider.version.project.node), timeout=15)
    resp.raise_for_status()
    result = resp.json()
    job_id = result.get("jobid")
    if not job_id:
        raise ValueError(f"爬虫启动失败：{result}")
    return job_id

def start_spiders(spiders: List[models.Spider]) -> bool:
    """批量启动爬虫"""
    for spider in spiders:
        start_spider(spider)
    return True


def stop_spider(spider: models.Spider) -> List[models.Job] | None:
    """停止某个爬虫的所有任务"""
    jobs = sync_jobs(spider.version.project.node)
    target_jobs = [j for j in jobs if j.spider == spider]
    stopped = []
    for job in target_jobs:
        stopped_job = stop_job(job)
        if stopped_job:
            stopped.append(stopped_job)
    return stopped or None


def stop_spiders(spiders: List[models.Spider]) -> List[models.Job] | None:
    """批量停止爬虫"""
    jobs = []
    for spider in spiders:
        job = stop_spider(spider)
        if job:
            jobs.extend(job)
    return jobs or None


def stop_job(job: models.Job) -> models.Job | None:
    """停止单个任务"""
    url = f"{job.node.url}/cancel.json"
    data = {
        "project": job.project.name,
        "job": job.job_id,
    }
    resp = requests.post(url, data=data, auth=_auth_for_node(job.node), timeout=15)
    resp.raise_for_status()
    result = resp.json()
    if result.get("status") == "ok":
        job.status = models.JobStatus.FINISHED
        job.end_time = timezone.now()
        job.save(update_fields=["status", "end_time", "update_time"])
        return job
    return None


def stop_jobs(jobs: List[models.Job]) -> List[models.Job]:
    """批量停止任务"""
    stopped_jobs = []
    for job in jobs:
        stopped_job = stop_job(job)
        if stopped_job:
            stopped_jobs.append(stopped_job)
    return stopped_jobs


def start_spider_group(group: SpiderGroupLike) -> List[models.Job]:
    """启动任务组里的所有爬虫"""
    spiders = group.resolved_spiders
    if not spiders:
        raise ValueError("group下面没有爬虫")
    job_ids = []
    for spider in spiders:
        job_id = start_spider(spider)
        job_ids.append(job_id)
    return job_ids


def stop_spider_group(group: models.SpiderGroup) -> List[models.Job]:
    """停止任务组里的所有爬虫"""
    jobs = []
    for spider in group.resolved_spiders:
        job = stop_spider(spider)
        if job:
            jobs.extend(job)
    return jobs


@django_ttl_cache()
def get_job_info(job: models.Job) -> dict:
    """获取某个任务的详细信息"""
    url = f"{job.node.url}/logs/{job.project.name}/{job.spider.name}/{job.job_id}.json"
    resp = requests.get(url, auth=_auth_for_node(job.node), timeout=15)
    resp.raise_for_status()
    return resp.json()


@django_ttl_cache()
def sync_jobs(node: models.Node) -> List[models.Job]:
    """列出节点上的所有任务并同步到数据库"""
    url = f"{node.url}/listjobs.json"
    jobs = []
    models.Job.objects.filter(node=node, status=models.JobStatus.PENDING).delete()
    for project in node.projects.all():
        resp = requests.get(url, params={"project": project.name}, auth=_auth_for_node(node), timeout=15)
        resp.raise_for_status()
        data = resp.json()
        for status, entries in data.items():
            if status not in ("pending", "running", "finished"):
                continue
            for entry in entries:
                spider_name = entry["spider"]
                start_time = entry.get("start_time")
                if start_time and settings.USE_TZ:
                    start_time = timezone.make_aware(datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S.%f"))
                else:
                    start_time = timezone.now()
                end_time = entry.get("end_time")
                if end_time and settings.USE_TZ:
                    end_time = timezone.make_aware(datetime.strptime(end_time, "%Y-%m-%d %H:%M:%S.%f"))
                job = models.Job(
                    node=node,
                    project=project,
                    spider=models.SpiderRegistry.objects.get(name=spider_name),
                    start_time=start_time,
                    job_id=entry["id"],
                    end_time=end_time,
                    items_url=entry.get("items_url"),
                    log_url=entry.get("log_url"),
                    pid=entry.get("pid"),
                    status=status,
                )
                job.gen_md5()
                jobs.append(job)

    models.Job.objects.bulk_create(jobs, ignore_conflicts=True)
    logger.info(f"{len(jobs)} jobs synced for {node}")
    return jobs


@django_ttl_cache()
def sync_project_versions(project: models.Project):
    url = f"{project.node.url}/listversions.json"
    resp = requests.get(url, params={"project": project.name}, auth=_auth_for_node(project.node), timeout=15)
    resp.raise_for_status()
    data = resp.json()
    versions = data.get("versions", [])
    project_versions = []
    for version in versions:
        project_versions.append(models.ProjectVersion(project=project, version=version, sync_status=models.SyncStatus.SUCCESS, scrapyd_exists=True))
    # models.ProjectVersion.objects.bulk_create(
    #     project_versions, update_conflicts=True, unique_fields=("project", "version"), update_fields=["sync_status", "scrapyd_exists"])
    # mysql不支持update_conflicts=True, 所以这里先插入, 再筛选更新
    models.ProjectVersion.objects.bulk_create(project_versions, ignore_conflicts=True)
    models.ProjectVersion.objects.filter(project=project, version__in=versions).update(sync_status=models.SyncStatus.SUCCESS, scrapyd_exists=True)
    models.ProjectVersion.objects.filter(~Q(version__in=versions), project=project).update(scrapyd_exists=False)
    logger.info(f"sync {len(project_versions)} versions for {project}")


@django_ttl_cache()
def sync_node_projects(node: models.Node, include_version=True):
    """列出某个节点上的项目，支持是否展开版本"""
    url = f"{node.url}/listprojects.json"
    resp = requests.get(url, auth=_auth_for_node(node), timeout=5)
    resp.raise_for_status()
    data = resp.json()
    scrapyd_projects = data.get("projects", [])
    projects = []
    for project_name in scrapyd_projects:
        project, created = models.Project.objects.update_or_create(node=node, name=project_name, defaults={"sync_status": models.SyncStatus.SUCCESS, "scrapyd_exists": True})
        projects.append(project)
        if include_version:
            sync_project_versions(project)
    models.Project.objects.filter(~Q(name__in=scrapyd_projects), node=node).update(scrapyd_exists=False)
    logger.info(f"sync {len(scrapyd_projects)} project for {node}")
    return projects


def sync_project_version_spiders(version: models.ProjectVersion) -> bool:
    """列出某个项目的爬虫"""
    if version.scrapyd_exists and (not version.is_spider_synced or version.spiders.count() == 0):
        url = f"{version.project.node.url}/listspiders.json"
        resp = requests.get(url, params={
            "project": version.project.name,
            "_version": version.version,
        }, auth=_auth_for_node(version.project.node), timeout=5)
        resp.raise_for_status()
        data = resp.json()
        spiders = data.get("spiders", [])
        spider_registries = [models.SpiderRegistry(name=spider) for spider in spiders]
        models.SpiderRegistry.objects.bulk_create(spider_registries, ignore_conflicts=True)

        results = [models.Spider(version=version, name=spider, registry_id=spider) for spider in spiders]
        models.Spider.objects.bulk_create(results)
        # 只需同步一次, 因为一个版本的spiders是不会变的
        logger.info(f"synced {len(spiders)} spiders for {version}@{version.project}")
        version.is_spider_synced = True
        version.save()
    return version.is_spider_synced


def add_version(version: models.ProjectVersion):
    """部署新版本"""
    url = f"{version.project.node.url}/addversion.json"
    if not version.egg_file:
        raise Exception("egg_file is not set")
    files = {"egg": version.egg_file.open()}
    data = {"project": version.project.name, "version": version.version}
    resp = requests.post(url, data=data, files=files, auth=_auth_for_node(version.project.node), timeout=15)
    resp.raise_for_status()
    return resp.json()


def delete_version(version: models.ProjectVersion):
    """删除某个版本"""
    url = f"{version.project.node.url}/delversion.json"
    data = {"project": version.project.name, "version": version.version}
    resp = requests.post(url, data=data, auth=_auth_for_node(version.project.node), timeout=15)
    resp.raise_for_status()
    ret = resp.json()
    if ret["status"] != "ok":
        raise ScrapydResponseError(ret["message"])


def delete_project(project: models.Project):
    """删除整个项目"""
    url = f"{project.node.url}/delproject.json"
    data = {"project": project.name}
    resp = requests.post(url, data=data, auth=_auth_for_node(project.node), timeout=15)
    resp.raise_for_status()
    ret = resp.json()
    if ret["status"] != "success":
        raise ScrapydResponseError(ret["message"])


def daemon_status(node: models.Node, timeout=3) -> dict:
    """获取节点的 daemon 状态"""
    url = f"{node.url}/daemonstatus.json"
    resp = requests.get(url, auth=_auth_for_node(node), timeout=timeout)
    resp.raise_for_status()
    return resp.json()


@django_ttl_cache()
def sync_nodes(with_jobs=False):
    nodes = models.Node.objects.all()
    available_nodes = []
    error_nodes = []
    for node in nodes:
        try:
            sync_node_projects(node, include_version=True)
            available_nodes.append(node)
        except Exception as e:
            logger.error(f"同步节点 {node} 失败: {e}")
            error_nodes.append(node)
    for version in models.ProjectVersion.objects.filter(project__node__in=available_nodes, scrapyd_exists=True, project__scrapyd_exists=True):
        sync_project_version_spiders(version)

    if with_jobs:
        for node in available_nodes:
            sync_jobs(node)

    if error_nodes:
        return f"节点{[node.name for node in available_nodes]}同步成功, 节点{[node.name for node in error_nodes]}同步失败"
    return ""






# Generated by Django 5.2.5 on 2025-09-12 15:43

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_scrapyd_manager', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='spidergroup',
            name='code',
            field=models.CharField(blank=True, max_length=100, null=True, validators=[django.core.validators.RegexValidator(code='invalid_code', message='代号只能包含英文字母', regex='^[a-zA-Z]*$')], verbose_name='代号'),
        ),
    ]



# scrapyd_manager/models.py
from __future__ import annotations
from datetime import datetime, timedelta
from django.core.validators import RegexValidator
from django.db import models
from django.utils import timezone
from django.utils.deconstruct import deconstructible
from .utils import get_md5
import json
import os


class Node(models.Model):
    name = models.CharField(max_length=100, verbose_name="节点名称", unique=True)
    ip = models.GenericIPAddressField(verbose_name="IP地址")
    port = models.IntegerField(default=6800, blank=True, null=True)
    ssl = models.BooleanField(default=False, verbose_name="是否启用SSL")
    description = models.CharField(max_length=500, blank=True, null=True, verbose_name="描述")
    auth = models.BooleanField(default=False, verbose_name="是否需要认证")
    username = models.CharField(max_length=255, blank=True, null=True)
    password = models.CharField(max_length=255, blank=True, null=True)
    create_time = models.DateTimeField(default=timezone.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    @classmethod
    def default_node(cls):
        for node in cls.objects.all().prefetch_related("projects"):
            if node.projects.count():
                return node
        return None

    @property
    def default_project(self):
        for project in self.projects.all():
            if project.versions.count() > 0:
                return project
        return None

    @classmethod
    def default_project_of_node(cls, node: str | int | Node):
        if isinstance(node, str):
            assert node.isdigit(), "node id非法 必须是整型"
        if not isinstance(node, Node):
            node = Node.objects.get(id=int(node))
        return node.default_project

    class Meta:
        ordering = ["-create_time"]
        db_table = "scrapyd_node"
        verbose_name = verbose_name_plural = "Scrapyd Node"

    def __str__(self):
        return self.name

    @property
    def url(self):
        host = self.ip or "localhost"
        port = self.port or 6800
        return f"{"https" if self.ssl else "http"}://{host}:{port}"


class SyncMode(models.TextChoices):
    AUTO = "auto", "自动"
    SYNC = "sync", "同步"
    NONE = "none", "无"
    

class SyncStatus(models.TextChoices):
    PENDING = "pending", "待同步"
    SUCCESS = "success", "成功"
    FAILED = "failed", "失败"


class Project(models.Model):
    node = models.ForeignKey(Node, on_delete=models.CASCADE, verbose_name="节点", db_constraint=False, related_name="projects")
    name = models.CharField(max_length=255, verbose_name="项目名")
    sync_mode = models.CharField(max_length=10, default=SyncMode.AUTO, choices=SyncMode.choices, verbose_name="同步模式")
    scrapyd_exists = models.BooleanField(default=False, verbose_name="在Scrapyd中存在")
    sync_status = models.CharField(max_length=10, default=SyncStatus.PENDING, choices=SyncStatus.choices, verbose_name="同步状态")
    create_time = models.DateTimeField(default=timezone.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    @property
    def latest_version(self):
        return self.versions.order_by("-version").first()

    class Meta:
        db_table = "scrapy_project"
        verbose_name = verbose_name_plural = "Scrapy Project"
        unique_together = (("node", "name"),)

    def __str__(self):
        return self.name


@deconstructible
class EggPath:
    def __call__(self, instance, filename):
        # 始终用项目名+版本号作为文件名
        base, ext = os.path.splitext(filename)
        return f"eggs/{instance.project.name}/{instance.version}{ext}"


class ProjectVersion(models.Model):
    project = models.ForeignKey(Project, on_delete=models.CASCADE, verbose_name="项目", db_constraint=False, related_name="versions")
    version = models.CharField(max_length=255, verbose_name='版本')
    is_spider_synced = models.BooleanField(default=False, verbose_name="是否已同步当前版本爬虫")
    egg_file = models.FileField(upload_to=EggPath(), null=True, blank=True, verbose_name="Egg 文件")
    description = models.CharField(max_length=200, null=True, blank=True, verbose_name="描述")
    sync_mode = models.CharField(max_length=10, choices=SyncMode.choices, default=SyncMode.AUTO, verbose_name="同步模式")
    scrapyd_exists = models.BooleanField(default=False, verbose_name="在Scrapyd中存在")
    sync_status = models.CharField(max_length=10, default=SyncStatus.PENDING, choices=SyncStatus.choices, verbose_name="同步状态")
    create_time = models.DateTimeField(default=timezone.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    @property
    def pretty(self):
        if self.version.isdigit():
            v = timezone.make_aware(datetime.fromtimestamp(int(self.version)))
        else:
            v = self.version
        return v

    @property
    def full_path(self):
        return f"{self.project.node.name}/{self.project.name}/{self.version}({self.pretty})"

    @property
    def short_path(self):
        pretty = self.pretty
        if pretty != self.version:
            return f"{self.version}({self.pretty})"
        return self.version

    class Meta:
        db_table = "scrapy_project_version"
        verbose_name = verbose_name_plural = "Scrapy Project Version"
        unique_together = (("project", "version"),)
        ordering = ["-version"]

    def __str__(self):
        return self.version


class SpiderRegistry(models.Model):
    name = models.CharField(max_length=255, verbose_name="爬虫名称", unique=True)
    description = models.CharField(max_length=200, null=True, blank=True, verbose_name="描述")
    kwargs = models.JSONField(default=dict, null=True, blank=True, verbose_name="Scrapy自定义参数(对组内所有爬虫生效)")
    settings = models.JSONField(default=dict, null=True, blank=True, verbose_name="Scrapy自定义设置(对组内所有爬虫生效)")
    create_time = models.DateTimeField(default=timezone.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    class Meta:
        db_table = "scrapy_spider_registry"
        verbose_name = verbose_name_plural = "Spider Registry"
        ordering = ["name"]

    def __str__(self):
        return self.name


class Spider(models.Model):
    registry = models.ForeignKey(SpiderRegistry, on_delete=models.DO_NOTHING, related_name="version_spiders", verbose_name="全局爬虫", to_field="name")
    version = models.ForeignKey(ProjectVersion, on_delete=models.CASCADE, verbose_name="版本", db_constraint=False, related_name="spiders")
    name = models.CharField(max_length=255, verbose_name="爬虫名称")
    kwargs = models.JSONField(default=dict, null=True, blank=True, verbose_name="Scrapy自定义参数(对组内所有爬虫生效)")
    settings = models.JSONField(default=dict, null=True, blank=True, verbose_name="Scrapy自定义设置(对组内所有爬虫生效)")
    create_time = models.DateTimeField(default=timezone.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    @property
    def fp(self):
        kwargs = self.kwargs.copy()
        for k, v in list(kwargs.items()):
            if k.startswith("__"):
                kwargs.pop(k)
        obj = {
            "version": self.version.version,
            "name": self.name,
            "kwargs": kwargs,
            "settings": self.settings,
        }
        s = json.dumps(obj, separators=(',', ':'), sort_keys=True)
        return get_md5(s)[:12]

    @property
    def job_id(self):
        group_code = self.kwargs.get("__group__") or "server"
        return f"{group_code}:{self.fp}:{self.version.version}:{timezone.now().strftime('%Y%m%d_%H%M%S')}"

    def __str__(self):
        return self.name

    class Meta:
        db_table = "scrapy_spider"
        verbose_name = verbose_name_plural = "Scrapy Spider"
        unique_together = (("version", "name"),)
        ordering = ["-version", "name"]


class SpiderGroup(models.Model):
    name = models.CharField(max_length=255, verbose_name="任务组名称", unique=True)
    code = models.CharField(max_length=100, null=True, blank=True, verbose_name="代号", validators=[
            RegexValidator(
                regex=r'^[a-zA-Z_-]*$',
                message='代号只能包含英文字母',
                code='invalid_code'
            )
        ])
    node = models.ForeignKey(Node, db_constraint=False, on_delete=models.DO_NOTHING, verbose_name="节点")
    project = models.ForeignKey(Project, db_constraint=False, verbose_name='项目', on_delete=models.CASCADE)
    version = models.ForeignKey(ProjectVersion, verbose_name='版本', null=True, blank=True, on_delete=models.CASCADE)
    spiders = models.ManyToManyField(SpiderRegistry, db_constraint=False, related_name="spiders", verbose_name="爬虫")
    kwargs = models.JSONField(default=dict, null=True, blank=True, verbose_name="Scrapy自定义参数kwargs(对组内所有爬虫生效)")
    settings = models.JSONField(default=dict, null=True, blank=True, verbose_name="Scrapy自定义设置settings(对组内所有爬虫生效)")
    description = models.CharField(max_length=200, blank=True, null=True, verbose_name="任务组描述")
    create_time = models.DateTimeField(default=timezone.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    @property
    def resolved_version(self):
        return self.version if self.version else self.project.latest_version

    @property
    def resolved_spiders(self):
        version = self.resolved_version
        registry_names = self.spiders.values_list("name", flat=True)
        spiders = Spider.objects.filter(
            version=version,
            version__project=self.project,
            name__in=registry_names
        )
        for spider in spiders:
            spider.kwargs["__group__"] = self.code
            spider.kwargs.update(self.kwargs)
            spider.settings.update(self.settings)
        return spiders

    class Meta:
        db_table = "scrapy_spider_group"
        verbose_name = verbose_name_plural = "Scrapy Spider Group"

    def __str__(self):
        return f"{self.node.name}/{self.project.name}/{self.name}"


class JobStatus(models.TextChoices):
    RUNNING = "running", "运行中"
    FINISHED = "finished", "已结束"
    PENDING = "pending", "启动中"


class Job(models.Model):
    node = models.ForeignKey(Node, on_delete=models.DO_NOTHING, verbose_name="节点", db_constraint=False, related_name="jobs")
    project = models.ForeignKey(Project, on_delete=models.DO_NOTHING, verbose_name="项目", db_constraint=False, related_name="jobs")
    spider = models.ForeignKey(SpiderRegistry, on_delete=models.DO_NOTHING, verbose_name="爬虫", db_constraint=False, related_name="jobs")
    version = models.CharField(max_length=200, verbose_name="版本", null=True, blank=True)
    job_id = models.CharField(max_length=255, verbose_name="任务ID")
    job_md5 = models.CharField(max_length=32, verbose_name="md5(job)", unique=True)
    start_time = models.DateTimeField(verbose_name="开始时间")
    end_time = models.DateTimeField(null=True, blank=True, verbose_name="结束时间")
    log_url = models.CharField(max_length=255, null=True, blank=True)
    items_url = models.CharField(max_length=255, null=True, blank=True)
    status = models.CharField(max_length=20, verbose_name="状态", choices=JobStatus.choices)
    pid = models.IntegerField(null=True, blank=True, verbose_name="进程ID")
    create_time = models.DateTimeField(default=timezone.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    def gen_md5(self):
        if not self.job_md5:
            start_time = self.start_time
            if isinstance(start_time, datetime):
                start_time = start_time.strftime("%Y-%m-%d %H:%M:%S.%f")
            fields = [self.project.name, self.spider.name, self.job_id, start_time]
            self.job_md5 = get_md5('-'.join(fields))
        return self.job_md5

    def save(self, *args, **kwargs):
        self.gen_md5()
        super().save(*args, **kwargs)

    @property
    def resolved_version(self):
        if not self.version:
            job_start_timestamp = int(self.start_time.timestamp())
            for version in self.project.versions.order_by("-version").all():
                if version.version.isdigit():
                    timestamp = int(version.version)
                    if timestamp < job_start_timestamp:
                        self.version = version
        return self.version

    class Meta:
        db_table = "scrapy_job"
        verbose_name = verbose_name_plural = "Scrapy Job"

    def __str__(self):
        return self.job_id


class JobInfoLog(models.Model):
    job = models.ForeignKey(Job, on_delete=models.DO_NOTHING, verbose_name="Job", db_constraint=False, related_name="logs")
    info = models.JSONField(null=True, blank=True, verbose_name="详情")
    create_time = models.DateTimeField(default=timezone.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    class Meta:
        db_table = "scrapy_job_info_log"
        verbose_name = verbose_name_plural = "Scrapy Job Info"

    def __str__(self):
        return str(self.job)


class GuardianStrategy(models.TextChoices):
    RESTART_ALWAYS = "restart_always", "始终重启"


class GuardianLock(models.Model):
    """
    用来保证同一时间只有一个 guardian 循环在运行
    """
    name = models.CharField(max_length=50, unique=True, default="default", verbose_name="锁名字")
    echo = models.IntegerField(default=0, verbose_name="运行轮次")
    guard_interval = models.IntegerField(default=60, verbose_name="监控周期(秒)")
    locked_at = models.DateTimeField(default=timezone.now, verbose_name="锁创建时间")
    heartbeat = models.DateTimeField(default=timezone.now, verbose_name="上一次心跳")
    create_time = models.DateTimeField(default=timezone.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    @property
    def is_expired(self):
        return self.expired_time < timezone.now()

    @property
    def expired_time(self):
        return self.heartbeat + timedelta(seconds=self.guard_interval * 2)

    class Meta:
        db_table = "scrapy_guardian_lock"
        verbose_name = verbose_name_plural = "Scrapy Guardian Lock"


class Guardian(models.Model):
    spider_group = models.ForeignKey(SpiderGroup, null=True, blank=True, on_delete=models.CASCADE, verbose_name="爬虫组", db_constraint=False)
    strategy = models.CharField(max_length=20, choices=GuardianStrategy.choices, default=GuardianStrategy.RESTART_ALWAYS, verbose_name="守护策略")
    description = models.CharField(max_length=200, null=True, blank=True, verbose_name="描述")
    enable = models.BooleanField(default=True, verbose_name="启用")
    interval = models.IntegerField(default=60, verbose_name="检测间隔(秒), 修改后重启生效")
    last_check = models.DateTimeField(null=True, blank=True, verbose_name="上次检测时间")
    last_action = models.CharField(max_length=255, null=True, blank=True, verbose_name="上次操作说明")
    create_time = models.DateTimeField(default=timezone.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    class Meta:
        db_table = "scrapy_guardian"
        verbose_name = verbose_name_plural = "Scrapy Guardian"

    def __str__(self):
        return f"Guardian[{self.spider_group}]"


class GuardianAction(models.TextChoices):
    PUBLISH_VERSION = "publish_version", "发布项目"
    START_SPIDER = "start_spider", "启动爬虫"


class GuardianLog(models.Model):
    guardian = models.ForeignKey(Guardian, on_delete=models.CASCADE, verbose_name="守护任务", db_constraint=False, related_name="logs")
    node = models.ForeignKey(Node, on_delete=models.CASCADE, verbose_name="节点")
    group = models.ForeignKey(SpiderGroup, on_delete=models.DO_NOTHING, db_constraint=False, verbose_name="关联爬虫组")
    spider = models.ForeignKey(Spider, null=True, blank=True, on_delete=models.DO_NOTHING, verbose_name="爬虫")
    spider_name = models.CharField(max_length=100, null=True, blank=True, verbose_name="爬虫名")
    action = models.CharField(max_length=100, choices=GuardianAction.choices, verbose_name="执行动作")
    reason = models.CharField(max_length=200, null=True, blank=True, verbose_name="原因")
    success = models.BooleanField(default=True, verbose_name="成功")
    message = models.TextField(null=True, blank=True, verbose_name="详细日志")
    create_time = models.DateTimeField(default=timezone.now, verbose_name="创建时间")

    class Meta:
        db_table = "scrapy_guardian_log"
        verbose_name = verbose_name_plural = "Scrapy Guardian Log"
        ordering = ["-create_time"]

    def __str__(self):
        return f"[{self.create_time}] {self.guardian} {self.action} ({self.success})"





import time
from functools import wraps
from django.core.cache import cache
from django.http import HttpRequest
from django.contrib.admin import ModelAdmin

_global_cache = {}
_sentinel = object()

_exclude_key_type = (ModelAdmin, HttpRequest)


def get_fun_cacheable_args_and(*args, **kwargs):
    key_args = []
    key_kwargs = {}
    for arg in args:
        if not isinstance(arg, _exclude_key_type):
            key_args.append(arg)
    for k, v in kwargs.items():
        if not isinstance(v, _exclude_key_type):
            key_kwargs[k] = v
    return key_args, key_kwargs


# 内存 TTL cache
def ttl_cache(ttl: int = 60):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            key_args, key_kwargs = get_fun_cacheable_args_and(*args, **kwargs)
            key = f"ttl_cache:{func.__name__}:args={key_args}:kwargs={key_kwargs}"

            now = time.time()
            if key in _global_cache:
                result, expire_at = _global_cache[key]
                if expire_at > now:
                    return result
            result = func(*args, **kwargs)
            _global_cache[key] = (result, now + ttl)
            return result
        return wrapper
    return decorator


def django_ttl_cache(ttl=10, prefix="ttl_cache"):
    import hashlib, pickle
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            key_args, key_kwargs = get_fun_cacheable_args_and(*args, **kwargs)
            raw_key = (func.__name__, key_args, key_kwargs)
            key = f"{prefix}:{hashlib.md5(pickle.dumps(raw_key)).hexdigest()}"

            result = cache.get(key, _sentinel)
            if result is not _sentinel:
                return result
            result = func(*args, **kwargs)
            cache.set(key, result, ttl)
            return result
        return wrapper
    return decorator



import time

from django import forms
from django.core.exceptions import ValidationError
from . import models
from .models import Project

ALLOWED_SCRAPY_CUSTOM_SETTINGS = {
  "CONCURRENT_REQUESTS": {
    "default": 16,
    "description": "Scrapy 全局最大并发请求数"
  },
  "CONCURRENT_REQUESTS_PER_DOMAIN": {
    "default": 8,
    "description": "每个域名最大并发请求数"
  },
  "CONCURRENT_REQUESTS_PER_IP": {
    "default": 0,
    "description": "每个 IP 最大并发请求数，0 表示禁用"
  },
  "DOWNLOAD_DELAY": {
    "default": 0,
    "description": "下载延迟，单位秒"
  },
  "RANDOMIZE_DOWNLOAD_DELAY": {
    "default": True,
    "description": "是否随机化下载延迟"
  },
  "AUTOTHROTTLE_ENABLED": {
    "default": False,
    "description": "是否启用自动限速"
  },
  "AUTOTHROTTLE_START_DELAY": {
    "default": 5,
    "description": "自动限速起始延迟"
  },
  "AUTOTHROTTLE_MAX_DELAY": {
    "default": 60,
    "description": "自动限速最大延迟"
  },
  "COOKIES_ENABLED": {
    "default": True,
    "description": "是否启用 Cookies"
  },
  "COOKIES_DEBUG": {
    "default": False,
    "description": "打印 Cookies 调试信息"
  },
  "RETRY_ENABLED": {
    "default": True,
    "description": "是否启用请求重试"
  },
  "RETRY_TIMES": {
    "default": 2,
    "description": "请求最大重试次数"
  },
  "DOWNLOAD_TIMEOUT": {
    "default": 180,
    "description": "下载超时，单位秒"
  },
  "REDIRECT_ENABLED": {
    "default": True,
    "description": "是否允许 HTTP 重定向"
  },
  "USER_AGENT": {
    "default": "Scrapy/VERSION (+https://scrapy.org)",
    "description": "请求 User-Agent"
  },
  "DEFAULT_REQUEST_HEADERS": {
    "default": {"Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8"},
    "description": "默认请求头"
  },
  "ITEM_PIPELINES": {
    "default": {},
    "description": "Item 管道配置，格式为 {'路径': 优先级}"
  },
  "DOWNLOADER_MIDDLEWARES": {
    "default": {},
    "description": "下载器中间件配置，格式为 {'路径': 优先级}"
  },
  "SPIDER_MIDDLEWARES": {
    "default": {},
    "description": "爬虫中间件配置，格式为 {'路径': 优先级}"
  },
  "LOG_LEVEL": {
    "default": "DEBUG",
    "description": "日志等级，可选 DEBUG/INFO/WARNING/ERROR/CRITICAL"
  },
  "LOG_ENABLED": {
    "default": True,
    "description": "是否启用日志"
  },
  "FEED_FORMAT": {
    "default": "json",
    "description": "导出文件格式，如 json/csv/xml"
  },
  "FEED_URI": {
    "default": "",
    "description": "导出文件路径或 URI"
  },
  "ROBOTSTXT_OBEY": {
    "default": True,
    "description": "是否遵守 robots.txt"
  },
  "DEPTH_LIMIT": {
    "default": 0,
    "description": "爬取深度限制，0 表示无限"
  },
  "DEPTH_PRIORITY": {
    "default": 0,
    "description": "深度优先或广度优先，>0 深度优先，<0 广度优先"
  },
  "EXTENSIONS": {
    "default": {},
    "description": "Scrapy 扩展配置，格式为 {'路径': 优先级}"
  },
  "TELNETCONSOLE_ENABLED": {
    "default": True,
    "description": "是否启用 Telnet 控制台"
  },
  "DUPEFILTER_CLASS": {
    "default": "scrapy.dupefilters.RFPDupeFilter",
    "description": "去重过滤器类"
  },
  "SCHEDULER": {
    "default": "scrapy.core.scheduler.Scheduler",
    "description": "Scrapy 调度器类"
  }
}


class SpiderGroupForm(forms.ModelForm):
    settings = forms.JSONField(
        required=False,
        help_text="Scrapy 支持的配置项(JSON)，示例: {\"CONCURRENT_REQUESTS\": 8}"
    )

    class Meta:
        model = models.SpiderGroup
        fields = "__all__"

    def __init__(self, *args, instance: models.SpiderGroup=None, **kwargs):
        super().__init__(*args, instance=instance, **kwargs)
        if instance is not None:
            self.fields["project"].widget.choices = [(p.pk, p.name) for p in instance.node.projects.all()]
            self.fields["version"].widget.choices = [(v.pk, v.version) for v in instance.project.versions.all()]
            self.fields["project"].queryset = models.Project.objects.filter(scrapyd_exists=True)
            self.fields["version"].queryset = models.ProjectVersion.objects.filter(scrapyd_exists=True)
        try:
            latest_version = instance.project.latest_version
        except AttributeError:
            latest_version = None
        if latest_version is None:
            version_empty_label = "自动最新版本[暂无可用版本]"
        else:
            version_empty_label = f"自动最新版本[{latest_version.short_path}]"
        self.fields["version"].empty_label = version_empty_label
        self.fields["version"].label_from_instance = lambda obj: obj.short_path

    def clean_settings(self):
        data = self.cleaned_data.get("settings") or {}
        if not isinstance(data, dict):
            raise ValidationError("custom_settings 必须是一个 JSON 对象")
        invalid_keys = [k for k in data.keys() if k not in ALLOWED_SCRAPY_CUSTOM_SETTINGS]
        if invalid_keys:
            raise ValidationError(f"无效的配置项: {invalid_keys}")
        return data

    def clean(self):
        cleaned_data = super().clean()
        if self.errors:
            return cleaned_data
        node = cleaned_data["node"]
        project = cleaned_data["project"]
        version: models.ProjectVersion = cleaned_data["version"]
        if project.node != node:
            raise ValidationError(f"project node({project.node})和group node({node})不一致！")
        if version:
            if version.project != project:
                raise ValidationError(f"version所输project{version.project}与group所输project({project})不一致！")
        return cleaned_data


class ProjectVersionForm(forms.ModelForm):
    project = forms.ModelChoiceField(queryset=Project.objects, required=True)
    node = forms.ModelChoiceField(queryset=models.Node.objects, required=True)
    egg_file = forms.FileField(
        required=True,
        label="Egg 文件上传",
        help_text="可手动上传 Scrapy 打包好的 egg 文件。"
    )
    version = forms.CharField(required=False)

    def __init__(self, *args, instance=None, **kwargs):
        super().__init__( *args, instance=instance, **kwargs)
        project: forms.ModelChoiceField = self.fields["project"] # type: ignore
        if instance is None:
            try:
                default_node = models.Node.default_node()
                self.fields["node"].initial = default_node
                project.widget.choices = [(p.pk, p.name) for p in default_node.projects.all()]
                project.initial = project.queryset.first()
            except models.Node.DoesNotExist:
                pass
        if self.instance.pk:
            self.instance: models.ProjectVersion
            self.fields["node"].initial = self.instance.project.node
            for name, field in self.fields.items():
                if name != "description":
                    field.disabled = True
                    field.required = False
        self.default_version = str(int(time.time()))
        self.fields["version"].help_text = self.default_version

    def clean(self):
        cleaned_data = super().clean()
        if self.errors:
            return cleaned_data
        if self.instance.pk:
            return {"description": cleaned_data["description"]}
        if not cleaned_data["version"]:
            cleaned_data["version"] = self.default_version
        node = cleaned_data.pop("node")
        if node != cleaned_data["project"].node:
            raise ValidationError("node和project所在node不一致")
        return cleaned_data

    class Meta:
        model = models.ProjectVersion
        fields = ["project", "version", "egg_file"]



import traceback
import time
from typing import Iterable
from django_scrapyd_manager import models, scrapyd_api
from django.utils import timezone
from django.db import transaction
from django.db.utils import IntegrityError
import logging


logger = logging.getLogger("django_scrapyd_manager")


class InvalidVersionError(Exception):
    pass


def node_has_project(node: models.Node, project: models.Project) -> bool:
    scrapyd_projects = scrapyd_api.sync_node_projects(node, include_version=False)
    for scrapyd_project in scrapyd_projects:
        if scrapyd_project == project:
            return True
    return False


def deploy_project_version(project_version: models.ProjectVersion):
    scrapyd_api.add_version(project_version)
    scrapyd_api.sync_project_version_spiders(project_version)


def missing_spiders_jobs_on_node(required_spiders: Iterable[models.Spider], node: models.Node) -> Iterable[models.Spider]:
    node_jobs = scrapyd_api.sync_jobs(node)
    job_ids = []
    missing_spiders = []
    for job in node_jobs:
        if job.status != models.JobStatus.FINISHED:
            job_ids.append(job.job_id)

    for required_spider in required_spiders:
        spider_fp = required_spider.fp
        for job_id in job_ids:
            if spider_fp in job_id:
                break
        else:
            missing_spiders.append(required_spider)
    return missing_spiders


def resolve_spiders_from_registries(registry_spiders: Iterable[models.SpiderRegistry], group: models.SpiderGroup) -> Iterable[models.Spider]:
    version = group.resolved_version
    resolved_spiders = version.spiders.filter(
        registry__in=registry_spiders,
    )
    resolved_spider_names = resolved_spiders.values_list("name", flat=True)
    registry_spider_names = [x.name for x in registry_spiders]
    for spider_name in registry_spider_names:
        if spider_name not in resolved_spider_names:
            raise RuntimeError(f"Spider {spider_name} not in {resolved_spider_names}")
    return resolved_spiders


class GuardSpiderGroup:

    def __init__(self, group: models.SpiderGroup, missing_spiders: Iterable[models.Spider]):
        self.resolved_spiders = missing_spiders
        self.name = group.name
        self.code = group.code
        self.kwargs = group.kwargs
        self.settings = group.settings


def acquire_guardian_lock(name="default") -> models.GuardianLock | None:
    now = timezone.now()
    try:
        with transaction.atomic():
            # 尝试创建新锁
            return models.GuardianLock.objects.create(name=name, locked_at=now, heartbeat=now)
    except IntegrityError:
        # 已经有锁，看看是否超时
        with transaction.atomic():
            lock = models.GuardianLock.objects.select_for_update().get(name=name)
            if lock.is_expired:
                # 锁过期 → 抢占
                lock.locked_at = now
                lock.heartbeat = now
                lock.save(update_fields=["locked_at", "heartbeat"])
                return lock
            expire_time = lock.expired_time
            last_heartbeat = lock.heartbeat
        # 非阻塞等待，看看 heartbeat 有没有更新
        while timezone.now() < expire_time:
            time.sleep(1)
            logger.debug(f"Guardian lock acquired at {timezone.now()}, expire at {expire_time}, waiting...")
            latest_heartbeat = models.GuardianLock.objects.get(pk=lock.pk).heartbeat
            if latest_heartbeat > last_heartbeat:
                # holder 还活着 → 放弃接管
                logger.info("Guardian lock heartbeat updated, another guardian is still running, exit.")
                return None

        logger.info("Guardian lock heartbeat timeout, trying to acquire lock...")
        # 超时，认为 holder 挂了 → 抢占
        with transaction.atomic():
            lock = models.GuardianLock.objects.select_for_update().get(name=name)
            if lock.heartbeat == last_heartbeat:  # 确认没更新
                lock.locked_at = now
                lock.heartbeat = now
                lock.save(update_fields=["locked_at", "heartbeat"])
                logger.info("Guardian lock acquired after timeout.")
                return lock
        return None


def refresh_guardian_lock(name="default"):
    models.GuardianLock.objects.filter(name=name).update(heartbeat=timezone.now())


def release_guardian_lock(name="default"):
    models.GuardianLock.objects.filter(name=name).delete()


def get_group_publishable_version(spider_group: models.SpiderGroup) -> models.ProjectVersion:
    version = spider_group.version
    if version is not None:
        if version.egg_file:
            return version
        else:
            raise InvalidVersionError(f"{spider_group.project}/{version.version}没有egg文件")
    version = spider_group.project.versions.filter(
        egg_file__isnull=False,
    ).order_by("-create_time", "-version").first()
    if version is None:
        raise InvalidVersionError(f"项目({spider_group.project})没有可用的带egg版本")
    return version


def guard_object(spider_guardian: models.Guardian):
    logs = []
    node = spider_guardian.spider_group.node
    if not node_has_project(node, spider_guardian.spider_group.project):
        log = models.GuardianLog(
            guardian=spider_guardian,
            node=node,
            group=spider_guardian.spider_group,
            action=models.GuardianAction.PUBLISH_VERSION,
            reason=f"node({node})上没有该项目:{spider_guardian.spider_group.project}"
        )
        try:
            version = get_group_publishable_version(spider_guardian.spider_group)
        except InvalidVersionError as e:
            log.success = False
            log.message = str(e)
        else:
            try:
                deploy_project_version(version)
            except Exception as e:
                log.success = False
                log.message = traceback.format_exc()
                logger.exception(e)
        log.save()
        logs.append(log)
    guard_spiders = spider_guardian.spider_group.resolved_spiders
    missing_spiders = missing_spiders_jobs_on_node(guard_spiders, node)

    if missing_spiders:
        for spider in missing_spiders:
            # 依次记录每个爬虫的启动状态
            log = models.GuardianLog(
                guardian=spider_guardian,
                node=node,
                spider=spider,
                spider_name=spider.name,
                group=spider_guardian.spider_group,
                action=models.GuardianAction.START_SPIDER,
                reason=f"{node}/{spider_guardian.spider_group.project}/{spider_guardian.spider_group.name}的爬虫{spider.name}没有运行"
            )
            try:
                guard_spider_group = GuardSpiderGroup(group=spider_guardian.spider_group, missing_spiders=[spider])
                scrapyd_api.start_spider_group(guard_spider_group)
            except Exception as e:
                log.success = False
                log.message = traceback.format_exc()
                logger.exception(e)
            log.save()
            logs.append(log)
    return logs


def guard_objects(objects: list[models.Guardian] = None):
    objects = objects or models.Guardian.objects.filter(enable=True).prefetch_related("spider_group", "spider_group__node", "spider_group__project")
    result_mapping = {}
    for obj in objects:
        name = (obj.description or "")[:20] or f"爬虫组守护{obj.spider_group.name}"
        try:
            logs = guard_object(obj)
            result_mapping[name] = {
                "success": True,
                "logs": logs
            }
            obj.last_action = logs[0].action if logs else "ok"
        except Exception as e:
            result_mapping[name] = {
                "success": False,
                "error": str(e)
            }
            logger.exception(e)
            obj.last_action = f"error: {e}"[:200]
        obj.last_check = timezone.now()
        obj.save()
    return result_mapping


# ANSI 颜色
COLORS = {
    "GREEN": "\033[32m",
    "RED": "\033[31m",
    "YELLOW": "\033[33m",
    "CYAN": "\033[36m",
    "RESET": "\033[0m",
}


def log_guard_results(result_mapping: dict, echo: int):
    """
    使用 logger 美化输出 Guardian 结果，终端显示彩色，文件输出普通文本。
    result_mapping 的 logs 是 GuardianLog queryset 或 list。
    """
    sep = "=" * 80
    logger.info(sep)
    logger.info(f"[Guardian Echo {echo}] Guarded {len(result_mapping)} objects")

    for name, result in result_mapping.items():
        success = result.get("success", False)
        if not success or result.get("logs"):
            status_text = "🔴 异常"
        else:
            status_text = "🟢 正常"
        color = COLORS["GREEN"] if success else COLORS["RED"]
        reset = COLORS["RESET"]

        # 输出任务组状态
        logger.info(f"-- {name}: {color}{status_text}{reset}")

        if success and "logs" in result:
            for log in result["logs"]:
                action = log.action
                spider_display = log.spider_name or (log.spider.name if log.spider else "")
                msg = log.message or ""
                s = "✅" if log.success else "❌"
                action_color = COLORS["CYAN"]
                success_color = COLORS["GREEN"] if log.success else COLORS["RED"]
                logger.info(f"   - [{action_color}{action}{reset}] {spider_display} {success_color}{s}{reset} {msg}")

        elif not success:
            error_msg = result.get("error", "Unknown error")
            logger.error(f"   - Error: {COLORS['RED']}{error_msg}{COLORS['RESET']}")

    logger.info(sep)


def guard_loop(lock_name="default"):
    lock = acquire_guardian_lock(lock_name)
    if lock is None:
        logger.warning("Another guardian is already running, exit.")
        return
    echo = 1
    try:
        while True:
            try:
                log_guard_results(guard_objects(), echo=echo)
                lock.heartbeat = timezone.now()
                lock.echo = echo
                lock.save(update_fields=["heartbeat", "echo"])
            except Exception as e:
                logger.exception("Guardian loop error: %s", e)
            echo += 1
            time.sleep(lock.guard_interval)
    finally:
        release_guardian_lock(lock_name)
        logger.info("Guardian loop stopped, lock released.")


from django.apps import AppConfig
import logging
import threading
import sys


class DjangoScrapydManagerConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'django_scrapyd_manager'
    verbose_name = "Scrapyd管理"
    _guardian_thread_started = False

    def ready(self):
        logger = logging.getLogger(self.name)
        if not logger.handlers:  # 避免重复添加
            handler = logging.StreamHandler()
            formatter = logging.Formatter("[%(levelname)s] %(asctime)s %(name)s %(message)s")
            handler.setFormatter(formatter)
            logger.addHandler(handler)
            logger.setLevel(logging.INFO)


        if "runserver" not in sys.argv and "gunicorn" not in sys.argv:
            return

        # 避免重复启动线程
        if self._guardian_thread_started:
            return

        from .guardian import guard_loop

        # 启动后台线程
        thread = threading.Thread(target=guard_loop, daemon=True)
        thread.start()
        self._guardian_thread_started = True
        logger.info("Guardian loop thread started")



# scrapyd_manager/admin.py
from functools import wraps

from django.contrib import admin, messages
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from datetime import datetime
from django.utils.html import format_html
from django.urls import path
from django.shortcuts import redirect
from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver
from django.urls import reverse
from django.utils.functional import lazy
from . import models
from . import scrapyd_api
from . import forms
import logging


admin_project_name = "django_scrapyd_manager"
# admin_prefix = settings.FORCE_SCRIPT_NAME if settings.FORCE_SCRIPT_NAME else ''


def _get_admin_index_url():
    url = reverse("admin:index")
    return url.rstrip('/')

admin_index_url = lazy(_get_admin_index_url, str)()


def _get_app_index_url():
    return f"{_get_admin_index_url()}/{admin_project_name}"

app_index_url = lazy(_get_app_index_url, str)()


logger = logging.getLogger("django_scrapyd_manager")


class ScrapydSyncAdminMixin:
    """
    通用 Mixin：给任何 ModelAdmin 自动加 scrapyd 同步
    - 同步频率由 @django_ttl_cache 控制
    """

    def sync_with_scrapyd(self):
        return scrapyd_api.sync_nodes()

    def _wrap_view(self, view):
        @wraps(view)
        def wrapper(request, *args, **kwargs):
            try:
                # 每次进入 admin 页面时都会触发
                error = self.sync_with_scrapyd()
                if error:
                    raise Exception(error)
                self.message_user(request, "同步Scrapyd数据成功", level=messages.SUCCESS)
            except Exception as e:
                logger.error(f"[scrapyd sync error]: {e}")
                self.message_user(request, f"同步Scrapyd数据失败: {e}", level=messages.ERROR)
            return view(request, *args, **kwargs)
        return wrapper

    def get_urls(self):
        urls = super().get_urls()
        for url in urls:
            url.callback = self._wrap_view(url.callback)  # 直接替换 view
        return urls


@receiver(post_delete, sender=models.Project)
def on_project_deleted(sender, instance: models.Project, **kwargs):
    if instance.sync_mode in (models.SyncMode.AUTO, models.SyncMode.SYNC):
        if instance.sync_status == models.SyncStatus.SUCCESS:
            try:
                scrapyd_api.delete_project(instance)
            except Exception as e:
                logger.exception(e)
                instance.sync_status = models.SyncStatus.FAILED
                instance.save()


@receiver(post_delete, sender=models.ProjectVersion)
def on_project_version_deleted(sender, instance: models.ProjectVersion, **kwargs):
    if instance.sync_mode in (models.SyncMode.AUTO, models.SyncMode.SYNC):
        if instance.sync_status == models.SyncStatus.SUCCESS:
            try:
                scrapyd_api.delete_version(instance)
            except Exception as e:
                logger.exception(e)
                instance.sync_status = models.SyncStatus.FAILED
                instance.save()


@receiver(post_save, sender=models.ProjectVersion)
def on_project_version_save(sender, instance: models.ProjectVersion, created, **kwargs):
    if instance.sync_mode != models.SyncMode.NONE:
        if instance.sync_status == models.SyncStatus.PENDING:
            try:
                scrapyd_api.add_version(instance)
                instance.sync_status = models.SyncStatus.SUCCESS
            except Exception as e:
                instance.sync_status = models.SyncStatus.FAILED
                logger.exception(e)
            instance.save(update_fields=["sync_status"])
    else:
        logger.info(f"{instance.sync_mode} is {instance.sync_status}, ignored.")


class CustomFilter(admin.SimpleListFilter):

    def choices(self, changelist):
        add_facets = changelist.add_facets
        facet_counts = self.get_facet_queryset(changelist) if add_facets else None
        for i, (lookup, title) in enumerate(self.lookup_choices):
            if add_facets:
                if (count := facet_counts.get(f"{i}__c", -1)) != -1:
                    title = f"{title} ({count})"
                else:
                    title = f"{title} (-)"
            yield {
                "selected": self.value() == str(lookup),
                "query_string": changelist.get_query_string(
                    {self.parameter_name: lookup}
                ),
                "display": title,
            }

    def queryset(self, request, queryset):
        value = self.value()
        if value:
            return queryset.filter(**{self.parameter_name: value})
        return queryset


@admin.register(models.Node)
class NodeAdmin(admin.ModelAdmin):
    list_display = ("name", "linked_url", "description", "related_projects", "auth", "daemon_status", "create_time")
    readonly_fields = ("create_time", "update_time")

    def linked_url(self, obj: models.Node) -> str:
        return format_html(f"<a href='{obj.url}'>{obj.url}</a>")
    linked_url.short_description = "Scrapyd地址"

    def daemon_status(self, obj: models.Node):
        try:
            return scrapyd_api.daemon_status(obj, timeout=0.5).get("status") == "ok"
        except Exception as e:
            logger.error(f"[scrapyd sync error]: {e}")
            return False
    daemon_status.short_description = "状态"
    daemon_status.boolean = True

    def related_projects(self, obj: models.Node):
        projects = []
        for project in obj.projects.all()[:5]:
            href = f"{app_index_url}/{models.Project._meta.model_name}/?node_id={obj.id}"
            projects.append(f"<a href='{href}'>{project.name}</a>")
        if len(projects) == 5:
            projects.append("···")
        return format_html('<span style="line-height: 1">%s</span>' % '<br>'.join(projects))
    related_projects.short_description = "项目"


class ProjectNodeFilter(CustomFilter):
    """右侧过滤：Node（节点）"""
    title = "节点"
    parameter_name = "node_id"

    def lookups(self, request, model_admin):
        return [(str(n.id), n.name) for n in models.Node.objects.all().order_by("name")]


class ProjectFilter(CustomFilter):
    """右侧过滤：Project（项目）"""
    title = "项目"
    parameter_name = "name"

    node_filter = ProjectNodeFilter

    def lookups(self, request, model_admin):
        node_id = request.GET.get(self.node_filter.parameter_name)
        if node_id:
            projects = models.Project.objects.filter(node_id=node_id).values_list("name", flat=True).distinct().order_by("name")
            return [(name, name) for name in projects]
        return []


@admin.register(models.Project)
class ProjectAdmin(ScrapydSyncAdminMixin, admin.ModelAdmin):
    list_display = ("node", "name", "latest_version", "related_versions", "scrapyd_exists", "sync_mode", "sync_status", "create_time")
    readonly_fields = ("sync_status", "create_time", "update_time")
    list_filter = (ProjectNodeFilter, )

    fields = (
        ("node", "sync_mode", "sync_status"),
        "name",
        "create_time",
        "update_time",
    )

    def has_change_permission(self, request, obj = ...):
        return False

    def delete_queryset(self, request, queryset):
        for obj in queryset:
            obj.delete()

    def latest_version(self, obj: models.Project):
        version = obj.latest_version
        if version:
            href = f"{app_index_url}/{models.ProjectVersion._meta.model_name}/?id={version.id}"
            return format_html(f'<a href="{href}">{version.pretty}</a>')
        return '-'
    latest_version.short_description = "最新版本"

    def related_versions(self, obj: models.Project):
        href = f"{app_index_url}/{models.ProjectVersion._meta.model_name}/?project_id={obj.id}"
        return format_html(f'<a href="{href}">{obj.versions.count()}</a>')
    related_versions.short_description = "版本数量"

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("node")


class VersionNodeFilter(ProjectNodeFilter):
    # 使用与 Django 内置 FieldListFilter 一致的参数名，便于复用已有的默认逻辑
    parameter_name = "project__node_id"


class VersionProjectFilter(ProjectFilter):
    """右侧过滤：Project（按项目名，不含版本），受 Node 选择联动"""
    parameter_name = "project__name"
    node_filter = VersionNodeFilter


@admin.register(models.ProjectVersion)
class ProjectVersionAdmin(ScrapydSyncAdminMixin, admin.ModelAdmin):
    list_display = ("id", "linked_version", "project", "spider_count", "has_egg_file", "description", "scrapyd_exists", "sync_mode", "sync_status", "is_spider_synced", "create_time")
    readonly_fields = ("is_spider_synced", "create_time", "update_time")
    list_filter = (VersionNodeFilter, VersionProjectFilter)
    form = forms.ProjectVersionForm
    fields = (
        ("node", "sync_mode", "sync_status"),
        "project",
        "version",
        "description",
        "egg_file",
    )

    def has_egg_file(self, obj: models.ProjectVersion):
        if obj.egg_file and obj.egg_file.storage.exists(obj.egg_file.name):
            return True
        return False
    has_egg_file.boolean = True
    has_egg_file.short_description = "egg文件"

    def linked_version(self, obj: models.ProjectVersion):
        if obj.version.isdigit():
            version_datetime = datetime.fromtimestamp(int(obj.version))
        else:
            version_datetime = obj.version
        href = f"{app_index_url}/{models.Spider._meta.model_name}/?version_id={obj.id}"
        return format_html(f'<a href="{href}">{obj.version}({version_datetime})</a>')
    linked_version.admin_order_field = "version"
    linked_version.short_description = "版本"

    def spider_count(self, obj: models.ProjectVersion):
        return obj.spiders.count()
    spider_count.short_description = "爬虫数量"

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("project", "project__node")

    class Media:
        js = ("admin/js/core.js", "admin/js/spider_group_linked.js")


class SpiderNodeFilter(ProjectNodeFilter):
    # 使用与 Django 内置 FieldListFilter 一致的参数名，便于复用已有的默认逻辑
    parameter_name = "version__project__node_id"


class SpiderProjectFilter(ProjectFilter):
    """右侧过滤：Project（按项目名，不含版本），受 Node 选择联动"""
    parameter_name = "version__project__name"
    node_filter = SpiderNodeFilter


class SpiderProjectVersionFilter(CustomFilter):
    title = "版本"
    parameter_name = "version"

    def lookups(self, request, model_admin):
        node_id = request.GET.get(SpiderNodeFilter.parameter_name)
        project_name = request.GET.get(SpiderProjectFilter.parameter_name)
        if not (project_name and node_id):
            return []
        # 获取当前节点下的所有项目版本
        versions = models.ProjectVersion.objects.filter(
            project__node_id=node_id,
            project__name=project_name,
        )
        return [(v.id, v.short_path) for v in versions]


@admin.register(models.SpiderRegistry)
class SpiderRegistryAdmin(ScrapydSyncAdminMixin, admin.ModelAdmin):
    list_display = ("name", "formatted_kwargs", "formatted_settings", "create_time")
    readonly_fields = ("name", "create_time", "update_time")

    def formatted_kwargs(self, obj: models.SpiderRegistry):
        args = []
        for key, value in obj.kwargs.items():
            args.append(f"{key} = {value}")
        if len(args) == 5:
            args.append("···")
        return format_html('<span style="line-height: 1">%s</span>' % '<br>'.join(args))
    formatted_kwargs.short_description = "参数(kwargs)"

    def formatted_settings(self, obj: models.SpiderRegistry):
        args = []
        for key, value in obj.settings.items():
            args.append(f"{key} = {value}")
        if len(args) == 5:
            args.append("···")
        return format_html('<span style="line-height: 1">%s</span>' % '<br>'.join(args))
    formatted_settings.short_description = "设置(setting)"

    def has_delete_permission(self, request, obj: models.Spider = None):
        return False


@admin.register(models.Spider)
class SpiderAdmin(ScrapydSyncAdminMixin, admin.ModelAdmin):
    list_display = ("name", "project_name", "project_node_name", "formatted_kwargs", "formatted_settings", "start_spider", "create_time")
    readonly_fields = ("version", "name", "create_time", "update_time")
    list_filter = (SpiderNodeFilter, SpiderProjectFilter, SpiderProjectVersionFilter)
    actions = ["start_spiders"]

    def formatted_kwargs(self, obj: models.SpiderGroup):
        args = []
        for key, value in obj.kwargs.items():
            args.append(f"{key} = {value}")
        if len(args) == 5:
            args.append("···")
        return format_html('<span style="line-height: 1">%s</span>' % '<br>'.join(args))
    formatted_kwargs.short_description = "参数(kwargs)"

    def formatted_settings(self, obj: models.SpiderGroup):
        args = []
        for key, value in obj.settings.items():
            args.append(f"{key} = {value}")
        if len(args) == 5:
            args.append("···")
        return format_html('<span style="line-height: 1">%s</span>' % '<br>'.join(args))
    formatted_settings.short_description = "设置(setting)"

    def has_delete_permission(self, request, obj: models.Spider = None):
        if obj is None:
            return False
        return True
        # return obj.version.sync_mode != models.SyncMode.NONE

    def get_urls(self):

        urls = super().get_urls()
        custom_urls = [
            path(
                "<path:spider_id>/start/",
                self.admin_site.admin_view(self.start_spider_view),
                name="scrapy_spider_start",
            ),
        ]
        return custom_urls + urls

    def start_spider_view(self, request, spider_id):
        spider = get_object_or_404(models.Spider, pk=spider_id)
        try:
            job_id = scrapyd_api.start_spider(spider)
            self.message_user(request, f"成功启动爬虫 {spider.name} (job_id={job_id})", level=messages.SUCCESS)
        except Exception as e:
            self.message_user(request, f"启动失败: {e}", level=messages.ERROR)
        from django.shortcuts import redirect
        return redirect(request.META.get("HTTP_REFERER", f"{app_index_url}/spider/"))

    def project_name(self, obj: models.Spider):
        return obj.version.project.name
    project_name.admin_order_field = "name"
    project_name.short_description = "项目名称"

    def project_node_name(self, obj: models.Spider):
        return obj.version.project.node.name
    project_node_name.admin_order_field = "project__node__name"
    project_node_name.short_description = "节点名称"

    def get_queryset(self, request):
        return super().get_queryset(request).prefetch_related("version", "version__project", "version__project__node")

    def start_spider(self, obj: models.Spider):
        return format_html(
            '<a class="button" href="{}">启动</a>',
            f"{app_index_url}/{models.Spider._meta.model_name}/{obj.id}/start/"
        )
    start_spider.short_description = "运行"

    def start_spiders(self, request, queryset):
        """启动选中的爬虫"""
        if not queryset:
            messages.error(request, "请选择要启动的爬虫")
            return
        for spider in queryset:
            try:
                job_id = scrapyd_api.start_spider(spider)
                messages.success(request, f"成功启动爬虫 {spider.name} (job_id={job_id})")
            except Exception as e:
                messages.error(request, f"启动爬虫 {spider.name} 失败: {str(e)}")
    start_spiders.short_description = "启动选中的爬虫"


@admin.register(models.SpiderGroup)
class SpiderGroupAdmin(ScrapydSyncAdminMixin, admin.ModelAdmin):
    list_display = ("name", "code", "node", "project", "related_spiders", "formatted_kwargs", "formatted_settings", "formatted_version", "start_spider_group", "create_time")
    readonly_fields = ("create_time", "update_time")
    filter_horizontal = ("spiders", )
    form = forms.SpiderGroupForm
    actions = ["start_group_spiders"]
    list_filter = ("node",)
    fields = (
        ("name", "code"),
        ("node", "project"),
        "version",
        "spiders",
        "description",
        "kwargs",
        "settings",
        "create_time",
        "update_time",
    )

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("project", "node", "version", "version__project", "version__project__node")

    def formatted_version(self, obj: models.SpiderGroup):
        if obj.version:
            return obj.version
        latest = obj.project.latest_version
        if latest:
            return f"自动最新版本[{latest.short_path}]"
        return "自动最新版本[暂无可用版本]"

    formatted_version.short_description = "版本"

    def formatted_kwargs(self, obj: models.SpiderGroup):
        args = []
        for key, value in obj.kwargs.items():
            args.append(f"{key} = {value}")
        if len(args) == 5:
            args.append("···")
        return format_html('<span style="line-height: 1">%s</span>' % '<br>'.join(args))
    formatted_kwargs.short_description = "参数(kwargs)"

    def formatted_settings(self, obj: models.SpiderGroup):
        args = []
        for key, value in obj.settings.items():
            args.append(f"{key} = {value}")
        if len(args) == 5:
            args.append("···")
        return format_html('<span style="line-height: 1">%s</span>' % '<br>'.join(args))
    formatted_settings.short_description = "设置(setting)"

    def related_spiders(self, obj: models.SpiderGroup):
        spiders = []
        for spider in obj.resolved_spiders:
            href = f"{app_index_url}/{models.Spider._meta.model_name}/?id={spider.id}"
            spiders.append(f"<a href='{href}'>{spider.name}</a>")
        if len(spiders) == 5:
            spiders.append("···")
        return format_html('<span style="line-height: 1">%s</span>' % '<br>'.join(spiders))
    related_spiders.short_description = "爬虫"

    def start_spider_group(self, obj: models.SpiderGroup):
        return format_html(
            '<a class="button" href="{}">启动</a>',
            f"{app_index_url}/{models.SpiderGroup._meta.model_name}/{obj.id}/start/"
        )
    start_spider_group.short_description = "运行"

    def start_group_spiders(self, request, queryset):
        """启动选中的爬虫组（组内所有爬虫）"""
        if not queryset:
            messages.error(request, "请选择要启动的爬虫组")
            return
        for group in queryset:
            try:
                scrapyd_api.start_spider_group(group)
                self.message_user(request, f"爬虫组{group.name} -> 启动成功", level=messages.SUCCESS)
            except Exception as e:
                self.message_user(request, f"爬虫组{group.name} -> 启动失败: {e}", level=messages.ERROR)
                break

    start_group_spiders.short_description = "启动选中的爬虫组"

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                "<path:group_id>/start/",
                self.admin_site.admin_view(self.start_group_view),
                name="scrapy_spidergroup_start",
            ),
            path(
                "api/node/<int:node_id>/projects/",
                self.admin_site.admin_view(self.get_projects),
                name="node_projects",
            ),
            path(
                "api/project/<int:project_id>/versions/",
                self.admin_site.admin_view(self.get_versions),
                name="project_versions",
            ),
            path(
                "api/version/spiders/",
                self.admin_site.admin_view(self.get_spiders),
                name="version_spiders",
            ),
        ]
        return custom_urls + urls

    @staticmethod
    def get_projects(request, node_id):
        projects = models.Project.objects.filter(node_id=node_id)
        data = [{"id": p.id, "text": p.name} for p in projects]
        return JsonResponse(data, safe=False)

    @staticmethod
    def get_versions(request, project_id):
        if project_id:
            versions = models.ProjectVersion.objects.filter(project_id=project_id).order_by("-create_time")
        else:
            node_id = request.GET.get("node_id")
            versions = models.ProjectVersion.objects.filter(project__node_id=node_id).order_by("-create_time")
        data = [{"id": v.id, "text": f"{v.version} ({v.pretty})"} for v in versions]
        return JsonResponse(data, safe=False)

    @staticmethod
    def get_spiders(request):
        version_id = request.GET.get("version_id")
        if version_id:
            spiders = models.Spider.objects.filter(version_id=version_id)
        else:
            node_id = request.GET.get("node_id")
            project_id = request.GET.get("project_id")
            spiders = models.Spider.objects.filter(version__project_id=project_id, version__project__node_id=node_id)
        data = [{"id": s.id, "text": s.name} for s in spiders]
        return JsonResponse(data, safe=False)

    def start_group_view(self, request, group_id):
        group = get_object_or_404(models.SpiderGroup, pk=group_id)
        try:
            scrapyd_api.start_spider_group(group)
            self.message_user(request, f"爬虫组{group.name} -> 启动成功", level=messages.SUCCESS)
        except Exception as e:
            self.message_user(request, f"爬虫组{group.name} -> 启动失败: {e}",
                              level=messages.ERROR)
        return redirect(request.META.get("HTTP_REFERER", f"{app_index_url}/{models.SpiderGroup._meta.model_name}/"))

    class Media:
        js = ("admin/js/core.js", f"admin/js/spider_group_linked.js")


class JobNodeFilter(ProjectNodeFilter):
    parameter_name = "node_id"


class JobProjectFilter(ProjectFilter):
    parameter_name = "project__name"
    node_filter = JobNodeFilter


class JobStatusFilter(CustomFilter):
    parameter_name = "status"
    title = "状态"

    def value(self):
        value = super().value()
        if not value:
            value = models.JobStatus.RUNNING
        return value

    def lookups(self, request, model_admin):
        filters = [
            (x, models.JobStatus[x.upper()].label) for x in models.Job.objects.values_list("status", flat=True).distinct()
        ]
        filters.sort()
        return filters


@admin.register(models.Job)
class JobAdmin(ScrapydSyncAdminMixin, admin.ModelAdmin):
    list_display = (
        "job_id", "job_spider", "job_project_version", "start_time", "end_time", "status", "pid", "job_sample_records", "job_info", "stop_job",
    )
    readonly_fields = ("create_time", "update_time", "start_time", "end_time", "pid", "log_url", "items_url", "spider", "status")
    list_filter = (JobStatusFilter, JobNodeFilter, JobProjectFilter)
    actions = ["stop_jobs"]
    ordering = ("-status", "-start_time")

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False

    def sync_with_scrapyd(self):
        return scrapyd_api.sync_nodes(with_jobs=True)

    def job_node(self, obj: models.Job):
        return obj.node.name
    job_node.admin_order_field = "node_id"
    job_node.short_description = "节点名称"

    def job_project(self, obj: models.Job):
        return obj.project.name
    job_project.admin_order_field = "project"
    job_project.short_description = "项目名称"

    def job_project_version(self, obj: models.Job):
        return obj.resolved_version
    job_project_version.admin_order_field = "project"
    job_project_version.short_description = "版本(根据job时间匹配)"

    def job_spider(self, obj: models.Job):
        return obj.spider.name
    job_spider.admin_order_field = "spider__name"
    job_spider.short_description = "爬虫名称"

    def job_sample_records(self, obj: models.Job):
        href = f"{app_index_url}/{models.JobInfoLog._meta.model_name}/?job={obj.id}"
        return format_html(f'<a class="button" href="{href}">采样记录({obj.logs.count()})</a>',)
    job_sample_records.short_description = "日志记录"

    def job_info(self, obj: models.Job):
        href = f"{app_index_url}/{models.JobInfoLog._meta.model_name}/?job={obj.id}"
        return format_html(f'<a class="button" href="{href}">Job最新状态</a>',)
    job_info.short_description = "Job最新状态"

    def get_list_display(self, request):
        if request.method != "GET":
            return self.list_display
        if request.GET.get("status") != "finished":
            list_display = list(self.list_display)
            list_display.remove("end_time")
            return list_display
        return self.list_display

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                "<path:job_id>/stop/",
                self.admin_site.admin_view(self.stop_job_view),
                name="scrapy_job_stop",
            ),
            path(
                "<path:job_id>/info/",
                self.admin_site.admin_view(self.sync_job_info_view),
                name="scrapy_job_stop",
            ),
        ]
        return custom_urls + urls

    def sync_job_info_view(self, request, job_id):
        job = get_object_or_404(models.Job, pk=job_id)
        try:
            info = scrapyd_api.get_job_info(job)
            models.JobInfoLog.objects.create(job=job, info=info)
            self.message_user(request, f"Job日志同步成功 {job.job_id} ({job.spider.name})", level=messages.SUCCESS)
        except Exception as e:
            self.message_user(request, f"Job日志同步失败: {e}", level=messages.ERROR)
        return redirect(request.META.get("HTTP_REFERER", f"{app_index_url}/{models.Job._meta.model_name}/"))

    def stop_job_view(self, request, job_id):
        job = get_object_or_404(models.Job, pk=job_id)
        try:
            scrapyd_api.stop_job(job)
            self.message_user(request, f"成功停止任务 {job.job_id} ({job.spider.name})", level=messages.SUCCESS)
        except Exception as e:
            self.message_user(request, f"停止任务失败: {e}", level=messages.ERROR)
        return redirect(request.META.get("HTTP_REFERER", f"{app_index_url}/{models.Job._meta.model_name}/"))

    def stop_job(self, obj: models.Job):
        if obj.status == models.JobStatus.RUNNING:  # 可选: 只在运行中显示按钮
            return format_html(
                '<a class="button" href="{}">停止</a>',
                f"{app_index_url}/{models.Job._meta.model_name}/{obj.id}/stop/"
            )
        return "-"
    stop_job.short_description = "操作"

    def stop_jobs(self, request, queryset):
        """停止选中的爬虫任务"""
        if not queryset:
            messages.error(request, "请选择要停止的任务")
            return
        for job in queryset:
            try:
                scrapyd_api.stop_job(job)
                messages.success(request, f"成功停止任务 {job.job_id} ({job.spider.name})")
            except Exception as e:
                messages.error(request, f"停止任务 {job.job_id} 失败: {str(e)}")
    stop_jobs.short_description = "停止选中的爬虫任务"

    def get_queryset(self, request):
        return super().get_queryset(request).prefetch_related("spider", "node", "project")

    def get_object(self, request, object_id, from_field = ...):
        return get_object_or_404(models.Job, pk=object_id)


@admin.register(models.JobInfoLog)
class JobInfoLogAdmin(admin.ModelAdmin):
    list_display = (
        "job", "job_node", "job_project",
    )

    list_filter = ("job__node", "job__project", "job__spider")
    ordering = ("-job_id", "-create_time")

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False

    def job_id(self, obj: models.JobInfoLog):
        return obj.job.job_id
    job_id.short_description = "JobId"

    def job_node(self, obj: models.JobInfoLog):
        return obj.job.node
    job_node.admin_order_field = "node_id"
    job_node.short_description = "节点名称"

    def job_project(self, obj: models.JobInfoLog):
        return obj.job.project
    job_project.admin_order_field = "project"
    job_project.short_description = "项目名称"

    def get_queryset(self, request):
        return super().get_queryset(request).prefetch_related("job")


@admin.register(models.GuardianLock)
class GuardianLockAdmin(admin.ModelAdmin):
    list_display = (
        "id", "name", "echo", "guard_interval", "heartbeat", "status", "locked_at", "create_time",
    )

    readonly_fields = (
        "name", "echo", "heartbeat", "locked_at", "create_time",
    )

    def status(self, obj: models.GuardianLock):
        return not obj.is_expired
    status.boolean = True
    status.short_description = "状态"


@admin.register(models.Guardian)
class GuardianAdmin(admin.ModelAdmin):
    list_display = (
        "id", "spider_group", "strategy", "description", "enable", "last_action", "interval", "last_check", "create_time",
    )
    list_editable = ("enable",)
    readonly_fields = ("last_action", "create_time", "update_time")
    fields = (
        ("spider_group", "strategy", "enable",),
        ("last_check", "last_action"),
        "interval",
        "create_time"
    )

    def get_queryset(self, request):
        return super().get_queryset(request).prefetch_related("spider_group", "spider_group__spiders")


@admin.register(models.GuardianLog)
class GuardianLogAdmin(admin.ModelAdmin):
    list_display = (
        "id", "guardian", "node", "spider_name", "action", "reason", "success", "create_time",
    )
    ordering = ("-create_time", )

    def has_change_permission(self, request, obj = ...):
        return False

    def has_add_permission(self, request):
        return False